#ifndef _ADATA_DEVICE_H
#define _ADATA_DEVICE_H

/*
** Device structure definition for the ADM-XRC series driver
*/

#include "platform.h"
#include "device_defs.h"
#include "plx.h"
#include "xplbridge.h"

#include "admxrc2/status.h"
#include "admxrc2/struct.h"

/*
** Declare some structures now so we can forward-reference them
*/
struct _DriverParams;
typedef struct _DriverParams DriverParams;
struct _ClientContext;
typedef struct _ClientContext ClientContext;
struct _DeviceContext;
typedef struct _DeviceContext DeviceContext;
struct _DmaQueue;
typedef struct _DmaQueue DmaQueue;
struct _DmaNode;
typedef struct _DmaNode DmaNode;
struct _DmaNodeLoadFpga;
typedef struct _DmaNodeLoadFpga DmaNodeLoadFpga;
struct _DmaTarget;
typedef struct _DmaTarget DmaTarget;
struct _DmaAbortContext;
typedef struct _DmaAbortContext DmaAbortContext;
struct _DmaScatterNode;
typedef struct _DmaScatterNode DmaScatterNode;
struct _DmaBounceBuffer;
typedef struct _DmaBounceBuffer DmaBounceBuffer;
struct _InterruptEvent;
typedef struct _InterruptEvent InterruptEvent;
struct _IntWaitNode;
typedef struct _IntWaitNode IntWaitNode;

typedef enum _ClockType {
  ClockType_16 = 0, /* 16.666MHz */
  ClockType_14 = 1  /* 14.318MHz */
} ClockType;

/*
** A structure representing a region of memory that has been
** locked down as a target for DMA transfers.
*/
struct _DmaTarget {
  unsigned long length;	/* size of mapped & locked region               */
  unsigned long	use;    /* (use > 0) => cannot be freed, is in use   */
  os_boolean bRemote;   /* TRUE => buf.remote is valid, FALSE => else buf.local */
  union {
    struct {
      OS_USBD* pUsbd;    /* description of mapped & locked region        */
      void* pBuffer;     /* user space address of mapped & locked region */
    } local;
    struct {
      os_uint32* pPagesPci;
      unsigned long initOffset;
      unsigned long pageLength;
      unsigned int pageBits;
    } remote;
  } variant;
};

/*
** A DMA completion callback routine - called when a DMA transfer
** is completed.
*/
#define DMACALLBACK_REASON_SUBMIT    (0)
#define DMACALLBACK_REASON_COLLECT   (1)
#define DMACALLBACK_REASON_COMPLETED (2)
typedef void (DmaCallback)(DeviceContext* pDevCtx, DmaNode* node);

/*
** A DMA queue, for queuing up DMA transfers
*/
struct _DmaQueue {
  int length;
  DmaNode* head;
  DmaNode* tail;
};

/*
** A context structure used when aborting DMA transfers
*/
struct _DmaAbortContext {
  /* initialised before call to atomic function */
  DeviceContext* devCtx;
  unsigned char  channel;

  /* filled in by call to atomic function */
  os_boolean     success;
};

/*
** A node in a scatter-gather list
*/
struct _DmaScatterNode {
  os_uint32  busAddr;
  os_uint32  localAddr;
  os_uint32  length;
  os_boolean bWrite;
};

/*
** A node in an Event list
*/
struct _InterruptEvent {
  os_event event;
  InterruptEvent* next;
};

/*
** A structure for controlling a DMA transfer; usually part of
** a queue of DMA transfers.
*/
struct _DmaNode {
  /* Client code must not touch these members */
  struct {
    DmaNode* prev;
    DmaNode* next;
    unsigned char state;
    unsigned char channel;
  } system;

  /* Client code must initialize these members */
  struct {
    void* context;  /* Client code can store some context information here */
    DmaCallback* callback; /* Called when client code must give attention */
    os_uint32 mode;
    unsigned long length;   /* WARNING - decremented during DMA */
    unsigned long pos;      /* WARNING - incremented during DMA */
    unsigned char direction;
    unsigned char queue;
  } client;

  /* Client code may inspect these members on completion */
  struct {
    int reason;   /* Reason for calling callback */
    OS_STATUS status;
  } status;
};

/*
** A structure for controlling an FPGA configuration operation that
** uses DMA. Has a DmaNode embedded in it.
*/
struct _DmaNodeLoadFpga {
  DmaNode dmanode;
  OS_EVENT event;
};

/*
** A structure for controlling a wait-for-interrupt operation.
*/
struct _IntWaitNode {
  OS_IO_REQUEST* pReq;
  OS_EVENT event;
  int state;
  os_boolean bCancel;
};

/*
** The driver parameters structure
*/
struct _DriverParams {
  os_boolean        pciReadMultiple;
  os_boolean        pciWriteAndInvalidate;
  os_uint32         maxDmaDesc;
  os_uint32         maxPlxDesc;
  os_boolean        pci9656DmaDisable;
  os_boolean        pci9656ErrataFix;
  os_boolean        pci9656RegsAsSpace;
  os_uint32         icd2061PMax;
  struct {
    struct {
      os_uint32     lclkMin, lclkMax;
    } xrc;
    struct {
      os_uint32     lclkMin, lclkMax;
    } xrcp;
    struct {
      os_uint32     lclkMin, lclkMax;
    } xrc2l;
    struct {
      os_uint32     lclkMin, lclkMax;
    } xrc2;
    struct {
      os_uint32     lclkMin, lclkMax;
    } wrc2;
    struct {
      os_uint32     lclkMin, lclkMax;
    } drc2;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    unlockFlash;
    } xpl;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    unlockFlash;
    } xp;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    unlockFlash;
    } xpi;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    smap32Bit;
    } xrc4ls;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    smap32Bit;
    } xrc4lx;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    smap32Bit;
    } xrc4sx;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    smap32Bit;
      os_boolean    unlockFlash;
    } xrc4fx;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    smap32Bit;
      os_boolean    unlockFlash;
    } adpedev;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    smap32Bit;
      os_boolean    unlockFlash;
    } xrc5lx;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    smap32Bit;
      os_boolean    unlockFlash;
    } xrc5t1;
    struct {
      os_uint32     lclkMin, lclkMax;
      os_boolean    smap32Bit;
      os_boolean    unlockFlash;
    } xrc5lxa;
  } model;
};

/*
** Client context structure; holds information about a handle that a user-space
** application has opened on a particular card.
*/

typedef enum _CoopLevel {
  CoopLevel_ReadOnly  = 0, /* Client not permitted to modify state of card, ie. can use information IOCTLs only */
  CoopLevel_Shared    = 1, /* Client permitted to modify state of card, and allows sharing of card with other clients */
  CoopLevel_Exclusive = 2  /* Client permitted to modify state of card, and disallows sharing of card with other clients */
} CoopLevel;

struct _ClientContext {
  OS_CLIENT_OBJECT* pClObj;    /* Link back to client object */
  CoopLevel coopLevel; /* Cooperative level of this handle */
  struct {
    void*         virtualBase;
    unsigned long virtualSize;
  } spaceInfo[ADMXRC_NUM_SPACE_MAX];
};

/*
** Definitions related to hardware methods.
*/

typedef enum _MethodResult {
  MethodResult_Success,            /* Successful                               */
  MethodResult_Timeout,            /* Timeout occurred during operation        */
  MethodResult_VerifyFailed,       /* Error verifying operation                */
  MethodResult_InvalidLocation,    /* Invalid location to access               */
  MethodResult_NotPresent,         /* Hardware is not present                  */
  MethodResult_NotSupported,       /* Method isn't supported for this hardware */
  MethodResult_InvalidParameter,   /* An invalid parameter was passed          */
  MethodResult_GeneralFailure      /* All other failures                       */
} MethodResult;

typedef MethodResult (MethodFlash)(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value);

typedef MethodResult (MethodFlashReset)(
  DeviceContext* pDevCtx);

typedef MethodResult (MethodIntEnab)(
  DeviceContext* pDevCtx,
  os_boolean     enable);

typedef os_boolean (MethodIsr)(
  DeviceContext* pDevCtx);

typedef MethodResult (MethodConfigMem)(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned long  location,
  os_uint32*     data);

typedef enum _MethodFpgaOp {
  MethodFpgaOp_SetProg,
  MethodFpgaOp_GetInit,
  MethodFpgaOp_GetDone
} MethodFpgaOp;

typedef MethodResult (MethodFpga)(
  DeviceContext* pDevCtx,
  MethodFpgaOp   op,
  void*          operand);

typedef MethodResult (MethodSelectMap)(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint8*      buffer,
  unsigned long  length);

typedef MethodResult (MethodConfigSpace)(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned int   space,
  os_uint32*     value);

typedef MethodResult (MethodAddDmaBuffer)(
  DeviceContext*  pDevCtx,
  unsigned int    chan,
  unsigned int    index,
  DmaScatterNode* snode);

typedef MethodResult (MethodDoDma)(
  DeviceContext*  pDevCtx,
  unsigned int    chan,
  os_boolean      start,
  os_uint32       mode,
  unsigned int    numBuffer);

typedef MethodResult (MethodClockWord)(
  DeviceContext*  pDevCtx,
  unsigned int    clockIndex,
  os_uint32       reqFreq,
  os_uint32*      actualFreq,
  os_uint32*      clockWord);

typedef MethodResult (MethodClockProg)(
  DeviceContext*  pDevCtx,
  unsigned int    index,
  os_uint32       actualFreq,
  os_uint32       clockWord,
  os_boolean      bBefore);

typedef MethodResult (MethodInitCardInfo)(
  DeviceContext*  pDevCtx);

typedef MethodResult (MethodLocalBusReset)(
  DeviceContext*  pDevCtx,
  os_boolean      bAssert);

typedef enum _DmaCompletionCode {
  DmaCompletionCode_Normal = 0,
  DmaCompletionCode_Failed = 1,
  DmaCompletionCode_EotRequested = 2
} DmaCompletionCode;

/*
** The device context structure
*/

struct _DeviceContext {
  /* Saved from onCreateDevice */
  OS_DRIVER_OBJECT*     pDrvObj;
  OS_DEVICE_OBJECT*     pDevObj;
  OS_DEVICE_ID          devId;
  OS_SLOT_ID            slotId;

  /* Resources given in onStartDevice */
  OS_DEVICE_RESOURCES   resources;

  /* Mapped/attached resources */
  OS_ATTACHED_IRQ*      irqObj;
  void*                 bar[6];

  /* Hardware mapped into kernel space */
  struct {
    union {
      void*             untyped;
      os_uint8*         byte;
      PLX9080REGS*      pci9080;
      PLX9656REGS*      pci9656;
      XPLBRIDGEREGS*    xpl;
      XRC4FXBRIDGEREGS* xrc4fx;
    } adapter;
    union {
      void*             untyped;
      os_uint8*         byte;
      CONTROLXRC*       plxbased;
      CONTROLWRC2*      wrc2;
      XPLSPACE1REGS*    xpl;
      XPSPACE1REGS*     xp;
      CONTROLXRC4LS*    xrc4ls;
      XRC4FXSPACE1REGS* xrc4fx;
      ADPEDEVSPACE1REGS* adpedev;
      XRC5LXSPACE1REGS* xrc5lx;
      XRC5LXASPACE1REGS* xrc5lxa;
    } control;
    void*               fpga;
    void*               space1;
  } hw;

  /* Shadows of important registers */
  struct {
    union {
      struct {
	os_uint32       intcsr;
	os_uint32       intcsr_mask;
	os_uint32       cntrl;
	os_uint16       dmacsr;
	os_uint8        imask;
      } plxbased;
      struct {
	os_uint32       pictl;
	os_uint32       fictl;
	os_uint32       dxctl[NUM_XPLBRIDGE_DMA_CHANNEL];
      } xpl;
      struct {
	os_uint32       pictl;
	os_uint32       fictl;
	os_uint32       dxctl[NUM_XPLBRIDGE_DMA_CHANNEL];
	os_uint32       fldata;
      } xp;
      struct {
	os_uint32       intcsr, intcsr_mask;
	os_uint32       ictl;
	os_uint32       flctl;
	os_uint32       fpctl;
	os_uint16       dmacsr;
	os_uint16       _pad1;
      } xrc4ls;
      struct {
	os_uint32       dxctl[NUM_XRC4FXBRIDGE_DMA_CHANNEL];
	os_uint32       pictl;
	os_uint32       fictl;
	os_uint32       flctl;
	os_uint32       fpctl;
      } xrc4fx;
      struct {
	os_uint32       pictl;
	os_uint32       fictl;
	os_uint32       dxctl[NUM_XPLBRIDGE_DMA_CHANNEL];
	os_uint32       flctl;
	os_uint32       fpctl;
      } adpedev;
    } regs;
    OS_SPINLOCK         lock;
  } shadow;

  struct {
    /* Are interrupts enabled or disabled? */
    os_boolean            enabled;
    OS_DPC*               fpgaintDpc;
    OS_DPC*               dmaintDpc[ADMXRC_NUM_DMA_CHANNEL_MAX];
  } interrupts;

  /* Tracks the open handle(s) to this device */
  struct {
    OS_CLIENT_OBJECT*   exclusive;
    unsigned long       numShared;
    OS_SPINLOCK         lock;
  } owner;

  /* DMA resources and data structures */
  struct {
    OS_SPINLOCK lock;
    OS_LCACHE* ctlPool;
    struct {
      OS_SG_MAP* sgMap;
      struct {
	void* real;                                    /* actual kernel address of descriptors */
	os_uint64 realPci;
	unsigned long realSize;
	union {
	  PLXDESC* plxbased;
	  XPLDMADESC* xpl;
	  void* untyped;
	} aligned;                                               /* aligned kernel address of descriptors */
	os_uint32 pci[ADMXRC_MAX_PLXDESC_PER_CHANNEL];     /* aligned PCI addresses of descriptors */
      } descriptors;
      unsigned long count;                                   /* Count of jobs on the DMA channel, should never be > 1 */ 
      DmaNode* waitsubmit;                              /* The job (if any) that is currently in the WAITBUFFERS state */
      DmaNode* active;                                  /* The job (if any) that is currently in the ACTIVE state */
      DmaNode* waitcollect;                             /* The job (if any) that is currently in the WAITCOLLECT state */
      unsigned long chunk;
      os_boolean cleanupPending;
      os_uint8 completionCode;                          /* Whether or not the DMA transfer was successful on the hardware */
    } channel[ADMXRC_NUM_DMA_CHANNEL_MAX];
    DmaQueue queue[ADMXRC_NUM_DMA_CHANNEL_MAX + 1];
    struct {
      unsigned int numChan;
      unsigned int descAlign;
      unsigned int descSize;
    } info;
  } dma;

  /* Memory regions mapped into user space */
  struct {
    OS_MAPPABLE_REGION* regions[ADMXRC_NUM_SPACE_MAX];
    int                 count;
  } userMapped;

  /* Card information, which user-mode threads can read */
  struct {
    struct {
      unsigned long cardId;
      os_uint32     serialNum;
      int           boardType;
      int           fpgaType;
      unsigned int  numClock;
      unsigned int  numDmaChan;
      unsigned int  numRamBank;
      unsigned int  numSpace;
      os_uint32     ramBanksFitted;
      os_uint8      boardRevision;
      os_uint8      logicRevision;
    } cardInfo;
    struct {
#if 0
      void*         virtualBase;
      unsigned long virtualSize;
#endif
      unsigned long physicalBase;
      unsigned long localBase;
      unsigned long localSize;
    } spaceInfo[ADMXRC_NUM_SPACE_MAX];
    struct {
      os_uint32     type;   /* bitmask of RAM mode(s) supported      */
      unsigned int  width;  /* in bits, valid whether fitted or not  */
      unsigned long size;   /* in words, valid whether fitted or not */
      os_boolean    fitted; /* TRUE if present (fitted to card)      */
    } bankInfo[ADMXRC_NUM_BANK_MAX];
    os_uint32           adapterLimit;
    ClockType           clockType;
    unsigned long       modifications; /* format is board-specific */
    struct {
      os_boolean        present;
      os_uint8          width;
      
      struct {
	char*           name;
	os_uint16       vendor, device;
      } identity;
      
      os_uint32         size;

      struct {
	unsigned int	numBootBlock;
	os_uint32       bootBlock[4];
	os_uint32       eraseBlock;
      } layout;
      
      os_uint8*         buffer;
      
      struct {
	os_uint32	start;
	os_uint32       length;
      } targetArea;
    } flash;

    os_uint32           spaceConfig[ADMXRC_NUM_SPACE_MAX];
  } info;

  struct {
    int                 numBar;
    os_uint32           selectMapOffset;
    os_boolean          bSelectMap32Bit;
  } miscInfo;

  /* FPGA information and control */
  struct {
    OS_LCACHE*          ctlPool;
    OS_SPINLOCK         lock;
    os_boolean          busy;
  } fpga;

  /* Interrupt waiting */
  struct {
    os_boolean          signalled;
    int                 state;
    IntWaitNode*        waiting;
    OS_EVENT            event;
    InterruptEvent*     eventList;
    OS_LCACHE*          ctlPool;
    OS_SPINLOCK         lock;
  } intwait;

  /* Clock generator programming */
  struct {
    OS_SPINLOCK         lock;
    os_boolean          busy;
    unsigned long       lclkMin, lclkMax;
    union {
      struct {
	os_uint32       lclk;
	os_uint32       mclk;
      } adpedev;
      struct {
	os_uint32       lclk;
	os_uint32       mclkA, mclkB;
      } xrc4fx;
      struct {
	os_uint32       lclk;
	os_uint32       mclk;
      } xrc5lx;
      struct {
	os_uint32       lclk;
	os_uint32       mclk;
      } xrc5lxa;
    } refClk;
  } clock;

  /* DMA descriptors */
  struct {
    DmaTarget**         table;     /* Maps DMADESCs to DmaTargets */
    int*                freestack; /* Stack of free table indices     */
    int                 top;       /* Stack pointer for 'freestack'   */
    OS_SPINLOCK         lock;
  } dmadesc;

  struct {
    MethodConfigMem*    configMem;
    MethodFlash*        flash;
    MethodFlashReset*   flashReset;
    MethodIntEnab*      intEnab;
    MethodIsr*          isr;
    MethodFpga*         fpga;
    MethodSelectMap*    selectMap;
    MethodConfigSpace*  configSpace;
    MethodAddDmaBuffer* addDmaBuffer;
    MethodDoDma*        doDma;
    MethodClockWord*    clockWord;
    MethodClockProg*    clockProg;
    MethodInitCardInfo* initCardInfo;
    MethodLocalBusReset* localBusReset;
  } methods;
};

#endif
