/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2006 Alpha Data
**
** dma.c - IOCTL handlers for the DMA API functions
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "drvparam.h"
#include "ioctl_struct.h"
#include "dma.h"
#include "dmadesc.h"
#include "dmacore.h"

/*
** This callback function is attached to the DMANODEs created by
** the DoDMA and DoDMAImmediate IOCTL handlers.
**
** May be called in an arbitrary context.
*/
static void
doDMACallback(
  DeviceContext*  devCtx,
  DmaNode* node)
{
  OS_EVENT* event;

  event = (OS_EVENT*) node->client.context;
  os_signal_event(event);
}

/*
** Handle ADMXRC_IOCTL_DODMA
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMA(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMA* iptr, * optr;
  DmaNode* dmanode = NULL;
  DmaTarget* dmatgt = NULL;
  OS_STATUS status = OS_STATUS_SUCCESS;
  unsigned long spanned;
  os_uint32 local;
  unsigned long length;
  os_uint32 mode;
  unsigned long flags;
  unsigned long channel;
  unsigned long dir;
  unsigned long offset;
  unsigned char queue;
  OS_EVENT event;
  unsigned long submit_len, submit_offs;
  os_boolean must_start;
  os_boolean bUseTimeout;
  os_ticks start, now, expire, ticks; /* ticks */
  unsigned long remaining_us; /* microseconds */
  os_boolean bRemote;

  iptr = (ADMXRC_IOCTLS_DODMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_DODMA*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC2_PCITOLOCAL && dir != ADMXRC2_LOCALTOPCI) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  local = iptr->in.local;
  length = iptr->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~ADMXRC2_DMAFLAG_DONOTQUEUE) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  dmanode = (DmaNode*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    return OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
  }

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = iptr->in.bUseTimeout;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(iptr->in.timeout);
  } else {
    start = expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  /* Lock down the user buffer and obtain an MDL for it */
  dmatgt = ADMXRC_GetDMATarget(devCtx, iptr->in.dmadesc);
  if (dmatgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
    goto cleanup;
  }
  bRemote = dmatgt->bRemote;

  /* Validate the offset and size parameters */
  offset = iptr->in.offset;
  if (offset >= dmatgt->length || (offset + length) > dmatgt->length) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? devCtx->info.cardInfo.numDmaChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMA: enqueuing DMANODE 0x%p dmadesc=%ld offset=%ld length=%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (unsigned long) iptr->in.dmadesc,
		  (unsigned long) offset,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  submit_offs = offset;
  submit_len = length;
  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_init_event(&event);

  must_start = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  while (1) {
    os_boolean bSignalled, bCancelled;

    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	ticks = 0;
      } else {
	ticks = expire - now;
      }
      bCancelled = os_wait_event_timeout(&event, ticks, TRUE, &bSignalled);
      if (bCancelled) {
	if (bSignalled) {
	  status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
	} else {
	  status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
	}
	if (!ADMXRC_CancelDma(devCtx, dmanode)) {
	  os_wait_event(&event, FALSE);
	  ADMXRC_CancelDma(devCtx, dmanode);
	}
	goto completed;
      }
    } else {
      bSignalled = os_wait_event(&event, TRUE);
      if (bSignalled) {
	status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
	if (!ADMXRC_CancelDma(devCtx, dmanode)) {
	  os_wait_event(&event, FALSE);
	  ADMXRC_CancelDma(devCtx, dmanode);
	}
	goto completed;
      }
    }

    switch (dmanode->status.reason) {
    case DMACALLBACK_REASON_SUBMIT:
      if (bRemote) {
	ADMXRC_SubmitDmaRemote(devCtx,
			       dmanode,
			       dmatgt->variant.remote.pPagesPci,
			       dmatgt->variant.remote.pageLength,
			       dmatgt->variant.remote.pageBits,
			       dmatgt->variant.remote.initOffset + submit_offs,
			       submit_len,
			       local,
			       &spanned);
      } else {
	ADMXRC_SubmitDmaUsbd(devCtx,
			     dmanode,
			     dmatgt->variant.local.pUsbd,
			     submit_offs,
			     submit_len,
			     local,
			     &spanned);
      }
      submit_offs += spanned;
      submit_len -= spanned;
      if (!(mode & ADMXRC2_DMAMODE_FIXEDLOCAL)) {
	local += spanned;
      }
      break;

    case DMACALLBACK_REASON_COLLECT:
      ADMXRC_CollectDma(devCtx, dmanode);
      break;

    case DMACALLBACK_REASON_COMPLETED:
      status = dmanode->status.status;
      goto completed;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      status = dmanode->status.status;
      goto completed;
    }
  }

 completed:
  if (OS_SUCCESS(status)) {
    req->status.length = sizeof(optr->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	optr->out.remaining = 0;
      } else {
	remaining_us = os_ticks_to_us(expire - now);
	if (remaining_us > iptr->in.timeout) {
	  optr->out.remaining = iptr->in.timeout;
	} else {
	  optr->out.remaining = remaining_us;
	}
      }
    }
  }

 cleanup:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }
  if (dmanode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMA2D
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMA2D(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMA2D* iptr, * optr;
  DmaNode* dmanode = NULL;
  DmaTarget* dmatgt = NULL;
  OS_STATUS status = OS_STATUS_SUCCESS;
  unsigned long spanned, local_spanned;
  os_uint32 local;
  unsigned long width;
  unsigned long skip, local_skip;
  unsigned long length;
  os_uint32 mode;
  unsigned long flags;
  unsigned long channel;
  unsigned long dir;
  unsigned long offset;
  unsigned char queue;
  OS_EVENT event;
  unsigned long submit_len, submit_offs;
  os_boolean must_start;
  os_boolean bUseTimeout;
  os_ticks start, now, expire, ticks; /* ticks */
  unsigned long remaining_us; /* microseconds */

  iptr = (ADMXRC_IOCTLS_DODMA2D*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_DODMA2D*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC2_PCITOLOCAL && dir != ADMXRC2_LOCALTOPCI) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  local = iptr->in.local;
  length = iptr->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate the 2D parameters */
  {
    unsigned long n_full_strip, n_skip;
    unsigned long partial_strip;

    width = iptr->in.width;
    skip = iptr->in.skip;
    local_skip = iptr->in.local_skip;

    n_full_strip = length / width;
    partial_strip = length % width;
    if (partial_strip) {
      n_skip = n_full_strip;
    } else {
      n_skip = n_full_strip - 1;
    }
    spanned = width * n_full_strip + partial_strip;
    local_spanned = spanned + n_skip * local_skip;
    spanned += n_skip * skip;
  }

  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~ADMXRC2_DMAFLAG_DONOTQUEUE) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  dmanode = (DmaNode*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    return OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
  }

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = iptr->in.bUseTimeout;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(iptr->in.timeout);
  } else {
    start = expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  /* Lock down the user buffer and obtain an MDL for it */
  dmatgt = ADMXRC_GetDMATarget(devCtx, iptr->in.dmadesc);
  if (dmatgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
    goto cleanup;
  }
  if (dmatgt->bRemote) {
    status = OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
    goto cleanup;
  }

  /* Validate the offset and size parameters */
  offset = iptr->in.offset;
  if (offset >= dmatgt->length || (offset + spanned) > dmatgt->length) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? devCtx->info.cardInfo.numDmaChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMA: enqueuing DMANODE 0x%p dmadesc=%ld offset=%ld length=%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (unsigned long) iptr->in.dmadesc,
		  (unsigned long) offset,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  submit_offs = offset;
  submit_len = length;
  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_init_event(&event);

  must_start = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  while (1) {
    unsigned long chunk;
    os_boolean bSignalled, bCancelled;

    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	ticks = 0;
      } else {
	ticks = expire - now;
      }
      bCancelled = os_wait_event_timeout(&event, ticks, TRUE, &bSignalled);
      if (bCancelled) {
	if (bSignalled) {
	  status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
	} else {
	  status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
	}
	if (ADMXRC_CancelDma(devCtx, dmanode)) {
	  goto completed;
	}
	os_wait_event(&event, FALSE);
      }
    } else {
      bSignalled = os_wait_event(&event, TRUE);
      if (bSignalled) {
	status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
	if (ADMXRC_CancelDma(devCtx, dmanode)) {
	  goto completed;
	}
	os_wait_event(&event, FALSE);
      }
    }

    switch (dmanode->status.reason) {
    case DMACALLBACK_REASON_SUBMIT:
      while (1) {
	chunk = (submit_len > width) ? width : submit_len;
	
	ADMXRC_SubmitDmaUsbd(devCtx, dmanode, dmatgt->variant.local.pUsbd, submit_offs, chunk, local, &spanned);
	if (!spanned) {
	  break;
	}
	submit_offs += (spanned + skip);
	submit_len -= spanned;
	if (!(mode & ADMXRC2_DMAMODE_FIXEDLOCAL)) {
	  local += (spanned + local_skip);
	}
      }
      break;

    case DMACALLBACK_REASON_COLLECT:
      ADMXRC_CollectDma(devCtx, dmanode);
      break;

    case DMACALLBACK_REASON_COMPLETED:
      status = dmanode->status.status;
      goto completed;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      status = dmanode->status.status;
      goto completed;
    }
  }

 completed:
  if (OS_SUCCESS(status)) {
    req->status.length = sizeof(optr->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	optr->out.remaining = 0;
      } else {
	remaining_us = os_ticks_to_us(expire - now);
	if (remaining_us > iptr->in.timeout) {
	  optr->out.remaining = iptr->in.timeout;
	} else {
	  optr->out.remaining = remaining_us;
	}
      }
    }
  }

 cleanup:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }
  if (dmanode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMAIMMEDIATE
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMAImmediate(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMAIMMEDIATE*	iptr, * optr;
  DmaNode* dmanode;
  DmaTarget* dmatgt = NULL;
  OS_STATUS status;
  unsigned long spanned;
  os_uint32 local;
  unsigned long length;
  os_uint32 mode;
  unsigned long flags;
  unsigned long channel;
  unsigned long dir;
  unsigned char queue;
  void* buffer;
  OS_EVENT event;
  unsigned long submit_len, submit_offs;
  os_boolean must_start;
  os_boolean bUseTimeout;
  os_ticks start, now, expire, ticks; /* ticks */
  unsigned long remaining_us; /* microseconds */

  iptr = (ADMXRC_IOCTLS_DODMAIMMEDIATE*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_DODMAIMMEDIATE*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC2_PCITOLOCAL && dir != ADMXRC2_LOCALTOPCI) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  buffer = iptr->in.buffer;
  local = iptr->in.local;
  length = iptr->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~ADMXRC2_DMAFLAG_DONOTQUEUE) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  dmanode = (DmaNode*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = iptr->in.bUseTimeout;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(iptr->in.timeout);
  } else {
    start = expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(devCtx, buffer, length, &dmatgt);
  if (!OS_SUCCESS(status)) {
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? devCtx->info.cardInfo.numDmaChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMAImmediate: enqueuing DMANODE 0x%p buffer=0x%p length=0x%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (void*) buffer,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  submit_offs = 0;
  submit_len = length;
  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_init_event(&event);

  must_start = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  while (1) {
    os_boolean bSignalled, bCancelled;

    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	ticks = 0;
      } else {
	ticks = expire - now;
      }
      bCancelled = os_wait_event_timeout(&event, ticks, TRUE, &bSignalled);
      if (bCancelled) {
	if (bSignalled) {
	  status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
	} else {
	  status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
	}
	if (!ADMXRC_CancelDma(devCtx, dmanode)) {
	  os_wait_event(&event, FALSE);
	  ADMXRC_CancelDma(devCtx, dmanode);
	}
	goto completed;
      }
    } else {
      bSignalled = os_wait_event(&event, TRUE);
      if (bSignalled) {
	status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
	if (!ADMXRC_CancelDma(devCtx, dmanode)) {
	  os_wait_event(&event, FALSE);
	  ADMXRC_CancelDma(devCtx, dmanode);
	}
	goto completed;
      }
    }

    switch (dmanode->status.reason) {
    case DMACALLBACK_REASON_SUBMIT:
      ADMXRC_SubmitDmaUsbd(devCtx, dmanode, dmatgt->variant.local.pUsbd, submit_offs, submit_len, local, &spanned);
      submit_offs += spanned;
      submit_len -= spanned;	
      if (!(mode & ADMXRC2_DMAMODE_FIXEDLOCAL)) {
	local += spanned;
      }
      break;

    case DMACALLBACK_REASON_COLLECT:
      ADMXRC_CollectDma(devCtx, dmanode);
      break;

    case DMACALLBACK_REASON_COMPLETED:
      status = dmanode->status.status;
      goto completed;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      status = dmanode->status.status;
      goto completed;
    }
  }

 completed:
  if (OS_SUCCESS(status)) {
    req->status.length = sizeof(optr->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	optr->out.remaining = 0;
      } else {
	remaining_us = os_ticks_to_us(expire - now);
	if (remaining_us > iptr->in.timeout) {
	  optr->out.remaining = iptr->in.timeout;
	} else {
	  optr->out.remaining = remaining_us;
	}
      }
    }
  }

 cleanup:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }
  if (dmanode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMAIMMEDIATE2D
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMAImmediate2D(
  DeviceContext* devCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_DODMAIMMEDIATE2D* iptr, * optr;
  DmaNode* dmanode;
  DmaTarget* dmatgt = NULL;
  OS_STATUS status;
  unsigned long spanned, local_spanned;
  os_uint32 local;
  unsigned long width;
  unsigned long skip, local_skip;
  unsigned long length;
  os_uint32 mode;
  unsigned long flags;
  unsigned long channel;
  unsigned long dir;
  unsigned char queue;
  void* buffer;
  OS_EVENT event;
  unsigned long submit_len, submit_offs;
  os_boolean must_start;
  os_boolean bUseTimeout;
  os_ticks start, now, expire, ticks; /* ticks */
  unsigned long remaining_us; /* microseconds */

  iptr = (ADMXRC_IOCTLS_DODMAIMMEDIATE2D*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_DODMAIMMEDIATE2D*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate PLX DMA mode */
  mode = iptr->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = iptr->in.direction;
  if (dir != ADMXRC2_PCITOLOCAL && dir != ADMXRC2_LOCALTOPCI) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  buffer = iptr->in.buffer;
  local = iptr->in.local;
  length = iptr->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate the 2D parameters */
  {
    unsigned long n_full_strip, n_skip;
    unsigned long partial_strip;

    width = iptr->in.width;
    skip = iptr->in.skip;
    local_skip = iptr->in.local_skip;

    n_full_strip = length / width;
    partial_strip = length % width;
    if (partial_strip) {
      n_skip = n_full_strip;
    } else {
      n_skip = n_full_strip - 1;
    }
    spanned = width * n_full_strip + partial_strip;
    local_spanned = spanned + n_skip * local_skip;
    spanned += n_skip * skip;
  }

  /* Validate flags parameter */
  flags = iptr->in.flags;
  if (flags & ~ADMXRC2_DMAFLAG_DONOTQUEUE) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate channel parameter */
  channel = iptr->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  dmanode = (DmaNode*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (dmanode == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = iptr->in.bUseTimeout;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(iptr->in.timeout);
  } else {
    start = expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(devCtx, buffer, spanned, &dmatgt);
  if (!OS_SUCCESS(status)) {
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? devCtx->info.cardInfo.numDmaChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  dmanode->client.context = (void*) &event;
  dmanode->client.callback = doDMACallback;
  dmanode->client.direction = (unsigned char) dir;
  dmanode->client.length = length;
  dmanode->client.pos = 0;
  dmanode->client.mode = mode;
  dmanode->client.queue = queue;

  os_dbgprint(6, ("ADMXRC_DoDMAImmediate: enqueuing DMANODE 0x%p buffer=0x%p length=0x%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*) dmanode,
		  (void*) buffer,
		  (unsigned long) length,
		  (unsigned long) local,
		  (unsigned long) dir,
		  (unsigned long) channel,
		  (unsigned long) mode,
		  (unsigned long) flags,
		  (unsigned long) queue));

  submit_offs = 0;
  submit_len = length;
  status = OS_STATUS_SUCCESS;
  req->user.context = (void*) dmanode;
  os_init_event(&event);

  must_start = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, dmanode, must_start)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  while (1) {
    unsigned long chunk;
    os_boolean bSignalled, bCancelled;

    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	ticks = 0;
      } else {
	ticks = expire - now;
      }
      bCancelled = os_wait_event_timeout(&event, ticks, TRUE, &bSignalled);
      if (bCancelled) {
	if (bSignalled) {
	  status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
	} else {
	  status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
	}
	if (ADMXRC_CancelDma(devCtx, dmanode)) {
	  goto completed;
	}
	os_wait_event(&event, FALSE);
      }
    } else {
      bSignalled = os_wait_event(&event, TRUE);
      if (bSignalled) {
	status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
	if (ADMXRC_CancelDma(devCtx, dmanode)) {
	  goto completed;
	}
	os_wait_event(&event, FALSE);
      }
    }

    switch (dmanode->status.reason) {
    case DMACALLBACK_REASON_SUBMIT:
      while (1) {
	chunk = (submit_len > width) ? width : submit_len;

	ADMXRC_SubmitDmaUsbd(devCtx, dmanode, dmatgt->variant.local.pUsbd, submit_offs, chunk, local, &spanned);
	if (!spanned) {
	  break;
	}
	submit_offs += (spanned + skip);
	submit_len -= spanned;	
	if (!(mode & ADMXRC2_DMAMODE_FIXEDLOCAL)) {
	  local += (spanned + local_skip);
	}
      }
      break;

    case DMACALLBACK_REASON_COLLECT:
      ADMXRC_CollectDma(devCtx, dmanode);
      break;

    case DMACALLBACK_REASON_COMPLETED:
      status = dmanode->status.status;
      goto completed;

    default:
      os_assert(FALSE);
      break;
    }

    os_reset_event(&event);
    if (ADMXRC_ContinueDma(devCtx, dmanode)) {
      status = dmanode->status.status;
      goto completed;
    }
  }

 completed:
  if (OS_SUCCESS(status)) {
    req->status.length = sizeof(optr->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	optr->out.remaining = 0;
      } else {
	remaining_us = os_ticks_to_us(expire - now);
	if (remaining_us > iptr->in.timeout) {
	  optr->out.remaining = iptr->in.timeout;
	} else {
	  optr->out.remaining = remaining_us;
	}
      }
    }
  }

 cleanup:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, dmatgt);
  }
  if (dmanode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, dmanode);
  }

  return status;
}
